"""Shape tests for broadcasting."""

from typing import Tuple, Type

import numpy as onp
from hypothesis import given, settings
from hypothesis import strategies as st
from jax import numpy as jnp
from utils import (
    assert_arrays_close,
    assert_transforms_close,
    general_group_test,
    sample_transform,
)

import jaxlie


@general_group_test
def test_broadcast_multiply(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    if batch_axes == ():
        return

    T = sample_transform(Group, batch_axes) @ sample_transform(Group)
    assert T.get_batch_axes() == batch_axes

    T = sample_transform(Group, batch_axes) @ sample_transform(Group, batch_axes=(1,))
    assert T.get_batch_axes() == batch_axes

    T = sample_transform(Group, batch_axes) @ sample_transform(
        Group, batch_axes=(1,) * len(batch_axes)
    )
    assert T.get_batch_axes() == batch_axes

    T = sample_transform(Group) @ sample_transform(Group, batch_axes)
    assert T.get_batch_axes() == batch_axes

    T = sample_transform(Group, batch_axes=(1,)) @ sample_transform(Group, batch_axes)
    assert T.get_batch_axes() == batch_axes


@general_group_test
def test_broadcast_apply(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    if batch_axes == ():
        return

    T = sample_transform(Group, batch_axes)
    points = onp.random.randn(Group.space_dim)
    assert (T @ points).shape == (*batch_axes, Group.space_dim)

    T = sample_transform(Group, batch_axes)
    points = onp.random.randn(1, Group.space_dim)
    assert (T @ points).shape == (*batch_axes, Group.space_dim)

    T = sample_transform(Group, batch_axes)
    points = onp.random.randn(*((1,) * len(batch_axes)), Group.space_dim)
    assert (T @ points).shape == (*batch_axes, Group.space_dim)

    T = sample_transform(Group)
    points = onp.random.randn(*batch_axes, Group.space_dim)
    assert (T @ points).shape == (*batch_axes, Group.space_dim)

    T = sample_transform(Group, batch_axes=(1,))
    points = onp.random.randn(*batch_axes, Group.space_dim)
    assert (T @ points).shape == (*batch_axes, Group.space_dim)



from typing import TYPE_CHECKING, Callable, Tuple, Type, TypeVar, Union, cast

import jax_dataclasses as jdc
from jax import numpy as jnp

from jaxlie.hints import Array

if TYPE_CHECKING:
    from .._base import MatrixLieGroup


T = TypeVar("T", bound="MatrixLieGroup")


def get_epsilon(dtype: jnp.dtype) -> float:
    """Helper for grabbing type-specific precision constants.

    Args:
        dtype: Datatype.

    Returns:
        Output float.
    """
    return {
        jnp.dtype("float32"): 1e-5,
        jnp.dtype("float64"): 1e-10,
    }[dtype]


def register_lie_group(
    *,
    matrix_dim: int,
    parameters_dim: int,
    tangent_dim: int,
    space_dim: int,
) -> Callable[[Type[T]], Type[T]]:
    """Decorator for registering Lie group dataclasses.

    Sets dimensionality class variables, and marks all methods for JIT compilation.
    """

    def _wrap(cls: Type[T]) -> Type[T]:
        # Register dimensions as class attributes.
        cls.matrix_dim = matrix_dim
        cls.parameters_dim = parameters_dim
        cls.tangent_dim = tangent_dim
        cls.space_dim = space_dim

        # JIT all methods.
        for f in filter(
            lambda f: not f.startswith("_")
            and callable(getattr(cls, f))
            and f != "get_batch_axes",  # Avoid returning tracers.
            dir(cls),
        ):
            setattr(cls, f, jdc.jit(getattr(cls, f)))

        return cls

    return _wrap


TupleOfBroadcastable = TypeVar(
    "TupleOfBroadcastable",
    bound="Tuple[Union[MatrixLieGroup, Array], ...]",
)


def broadcast_leading_axes(inputs: TupleOfBroadcastable) -> TupleOfBroadcastable:
    """Broadcast leading axes of arrays. Takes tuples of either:
    - an array, which we assume has shape (*, D).
    - a Lie group object."""

    from .._base import MatrixLieGroup

    array_inputs = [
        (
            (x.parameters(), (x.parameters_dim,))
            if isinstance(x, MatrixLieGroup)
            else (x, x.shape[-1:])
        )
        for x in inputs
    ]
    for array, shape_suffix in array_inputs:
        assert array.shape[-len(shape_suffix) :] == shape_suffix
    batch_axes = jnp.broadcast_shapes(
        *[array.shape[: -len(suffix)] for array, suffix in array_inputs]
    )
    broadcasted_arrays = tuple(
        jnp.broadcast_to(array, batch_axes + shape_suffix)
        for (array, shape_suffix) in array_inputs
    )
    return cast(
        TupleOfBroadcastable,
        tuple(
            array if not isinstance(inp, MatrixLieGroup) else type(inp)(array)
            for array, inp in zip(broadcasted_arrays, inputs)
        ),
    )



"""Tests for general operation definitions."""

from typing import Tuple, Type

import numpy as onp
from hypothesis import given, settings
from hypothesis import strategies as st
from jax import numpy as jnp
from utils import (
    assert_arrays_close,
    assert_transforms_close,
    general_group_test,
    sample_transform,
)

import jaxlie


@general_group_test
def test_sample_uniform_valid(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that sample_uniform() returns valid group members."""
    T = sample_transform(Group, batch_axes)  # Calls sample_uniform under the hood.
    assert_transforms_close(T, T.normalize())


@settings(deadline=None)
@given(_random_module=st.random_module())
def test_so2_from_to_radians_bijective(_random_module):
    """Check that we can convert from and to radians."""
    radians = onp.random.uniform(low=-onp.pi, high=onp.pi)
    assert_arrays_close(jaxlie.SO2.from_radians(radians).as_radians(), radians)


@settings(deadline=None)
@given(_random_module=st.random_module())
def test_so3_xyzw_bijective(_random_module):
    """Check that we can convert between xyzw and wxyz quaternions."""
    T = sample_transform(jaxlie.SO3)
    assert_transforms_close(T, jaxlie.SO3.from_quaternion_xyzw(T.as_quaternion_xyzw()))


@settings(deadline=None)
@given(_random_module=st.random_module())
def test_so3_rpy_bijective(_random_module):
    """Check that we can convert between quaternions and Euler angles."""
    T = sample_transform(jaxlie.SO3)
    assert_transforms_close(T, jaxlie.SO3.from_rpy_radians(*T.as_rpy_radians()))


@general_group_test
def test_log_exp_bijective(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check 1-to-1 mapping for log <=> exp operations."""
    transform = sample_transform(Group, batch_axes)

    tangent = transform.log()
    assert tangent.shape == (*batch_axes, Group.tangent_dim)

    exp_transform = Group.exp(tangent)
    assert_transforms_close(transform, exp_transform)
    assert_arrays_close(tangent, exp_transform.log())


@general_group_test
def test_inverse_bijective(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check inverse of inverse."""
    transform = sample_transform(Group, batch_axes)
    assert_transforms_close(transform, transform.inverse().inverse())


@general_group_test
def test_matrix_bijective(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that we can convert to and from matrices."""
    transform = sample_transform(Group, batch_axes)
    assert_transforms_close(transform, Group.from_matrix(transform.as_matrix()))


@general_group_test
def test_adjoint(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check adjoint definition."""
    transform = sample_transform(Group, batch_axes)
    omega = onp.random.randn(*batch_axes, Group.tangent_dim)
    assert_transforms_close(
        transform @ Group.exp(omega),
        Group.exp(onp.einsum("...ij,...j->...i", transform.adjoint(), omega))
        @ transform,
    )


@general_group_test
def test_repr(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Smoke test for __repr__ implementations."""
    transform = sample_transform(Group, batch_axes)
    print(transform)


@general_group_test
def test_apply(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check group action interfaces."""
    T_w_b = sample_transform(Group, batch_axes)
    p_b = onp.random.randn(*batch_axes, Group.space_dim)

    if Group.matrix_dim == Group.space_dim:
        assert_arrays_close(
            T_w_b @ p_b,
            T_w_b.apply(p_b),
            onp.einsum("...ij,...j->...i", T_w_b.as_matrix(), p_b),
        )
    else:
        # Homogeneous coordinates.
        assert Group.matrix_dim == Group.space_dim + 1
        assert_arrays_close(
            T_w_b @ p_b,
            T_w_b.apply(p_b),
            onp.einsum(
                "...ij,...j->...i",
                T_w_b.as_matrix(),
                onp.concatenate([p_b, onp.ones_like(p_b[..., :1])], axis=-1),
            )[..., :-1],
        )


@general_group_test
def test_multiply(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check multiply interfaces."""
    T_w_b = sample_transform(Group, batch_axes)
    T_b_a = sample_transform(Group, batch_axes)
    assert_arrays_close(
        onp.einsum(
            "...ij,...jk->...ik", T_w_b.as_matrix(), T_w_b.inverse().as_matrix()
        ),
        onp.broadcast_to(
            onp.eye(Group.matrix_dim), (*batch_axes, Group.matrix_dim, Group.matrix_dim)
        ),
    )
    assert_arrays_close(
        onp.einsum(
            "...ij,...jk->...ik", T_w_b.as_matrix(), jnp.linalg.inv(T_w_b.as_matrix())
        ),
        onp.broadcast_to(
            onp.eye(Group.matrix_dim), (*batch_axes, Group.matrix_dim, Group.matrix_dim)
        ),
    )
    assert_transforms_close(T_w_b @ T_b_a, Group.multiply(T_w_b, T_b_a))



from ._backprop import grad as grad
from ._backprop import value_and_grad as value_and_grad
from ._backprop import zero_tangents as zero_tangents
from ._deltas import rminus as rminus
from ._deltas import rplus as rplus
from ._deltas import (
    rplus_jacobian_parameters_wrt_delta as rplus_jacobian_parameters_wrt_delta,
)
from ._tree_utils import normalize_all as normalize_all



"""Test transform serialization, for things like saving calibrated transforms to
disk."""

from typing import Tuple, Type

import flax.serialization
from utils import assert_transforms_close, general_group_test, sample_transform

import jaxlie


@general_group_test
def test_serialization_state_dict_bijective(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check bijectivity of state dict representation conversions."""
    T = sample_transform(Group, batch_axes)
    T_recovered = flax.serialization.from_state_dict(
        T, flax.serialization.to_state_dict(T)
    )
    assert_transforms_close(T, T_recovered)


@general_group_test
def test_serialization_bytes_bijective(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check bijectivity of byte representation conversions."""
    T = sample_transform(Group, batch_axes)
    T_recovered = flax.serialization.from_bytes(T, flax.serialization.to_bytes(T))
    assert_transforms_close(T, T_recovered)



from __future__ import annotations

from typing import Any, Callable, Sequence, Tuple, Union, overload

import jax
from jax import numpy as jnp
from typing_extensions import ParamSpec

from .._base import MatrixLieGroup
from . import _deltas, _tree_utils


def zero_tangents(pytree: Any) -> _tree_utils.TangentPytree:
    """Replace all values in a Pytree with zero vectors on the corresponding tangent
    spaces."""

    def tangent_zero(t: MatrixLieGroup) -> jax.Array:
        return jnp.zeros(t.get_batch_axes() + (t.tangent_dim,))

    return _tree_utils._map_group_trees(
        tangent_zero,
        lambda array: jnp.zeros_like(array),
        pytree,
    )


AxisName = Any

P = ParamSpec("P")


@overload
def grad(
    fun: Callable[P, Any],
    argnums: int = 0,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: Sequence[AxisName] = (),
) -> Callable[P, _tree_utils.TangentPytree]: ...


@overload
def grad(
    fun: Callable[P, Any],
    argnums: Sequence[int],
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: Sequence[AxisName] = (),
) -> Callable[P, Tuple[_tree_utils.TangentPytree, ...]]: ...


def grad(
    fun: Callable[P, Any],
    argnums: Union[int, Sequence[int]] = 0,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: Sequence[AxisName] = (),
):
    """Same as `jax.grad`, but computes gradients of Lie groups with respect to
    tangent spaces."""

    compute_value_and_grad = value_and_grad(
        fun=fun,
        argnums=argnums,
        has_aux=has_aux,
        holomorphic=holomorphic,
        allow_int=allow_int,
        reduce_axes=reduce_axes,
    )

    def grad_fun(*args, **kwargs):
        ret = compute_value_and_grad(*args, **kwargs)
        if has_aux:
            return ret[1], ret[0][1]
        else:
            return ret[1]

    return grad_fun


@overload
def value_and_grad(
    fun: Callable[P, Any],
    argnums: int = 0,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: Sequence[AxisName] = (),
) -> Callable[P, Tuple[Any, _tree_utils.TangentPytree]]: ...


@overload
def value_and_grad(
    fun: Callable[P, Any],
    argnums: Sequence[int],
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: Sequence[AxisName] = (),
) -> Callable[P, Tuple[Any, Tuple[_tree_utils.TangentPytree, ...]]]: ...


def value_and_grad(
    fun: Callable[P, Any],
    argnums: Union[int, Sequence[int]] = 0,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: Sequence[AxisName] = (),
):
    """Same as `jax.value_and_grad`, but computes gradients of Lie groups with respect to
    tangent spaces."""

    def wrapped_grad(*args, **kwargs):
        def tangent_fun(*tangent_args, **tangent_kwargs):
            return fun(  # type: ignore
                *_deltas.rplus(args, tangent_args),
                **_deltas.rplus(kwargs, tangent_kwargs),
            )

        # Put arguments onto tangent space.
        tangent_args = map(zero_tangents, args)
        tangent_kwargs = {k: zero_tangents(v) for k, v in kwargs.items()}

        return jax.value_and_grad(
            fun=tangent_fun,
            argnums=argnums,
            has_aux=has_aux,
            holomorphic=holomorphic,
            allow_int=allow_int,
            reduce_axes=reduce_axes,
        )(*tangent_args, **tangent_kwargs)

    return wrapped_grad  # type: ignore



from typing import Tuple, cast

import jax
import jax_dataclasses as jdc
from jax import numpy as jnp
from typing_extensions import override

from . import _base, hints
from ._so2 import SO2
from .utils import broadcast_leading_axes, get_epsilon, register_lie_group


@register_lie_group(
    matrix_dim=3,
    parameters_dim=4,
    tangent_dim=3,
    space_dim=2,
)
@jdc.pytree_dataclass
class SE2(_base.SEBase[SO2]):
    """Special Euclidean group for proper rigid transforms in 2D. Broadcasting
    rules are the same as for numpy.

    Internal parameterization is `(cos, sin, x, y)`. Tangent parameterization is `(vx,
    vy, omega)`.
    """

    # SE2-specific.

    unit_complex_xy: jax.Array
    """Internal parameters. `(cos, sin, x, y)`. Shape should be `(*, 3)`."""

    @override
    def __repr__(self) -> str:
        unit_complex = jnp.round(self.unit_complex_xy[..., :2], 5)
        xy = jnp.round(self.unit_complex_xy[..., 2:], 5)
        return f"{self.__class__.__name__}(unit_complex={unit_complex}, xy={xy})"

    @staticmethod
    def from_xy_theta(x: hints.Scalar, y: hints.Scalar, theta: hints.Scalar) -> "SE2":
        """Construct a transformation from standard 2D pose parameters.

        Note that this is not the same as integrating over a length-3 twist.
        """
        cos = jnp.cos(theta)
        sin = jnp.sin(theta)
        return SE2(unit_complex_xy=jnp.stack([cos, sin, x, y], axis=-1))

    # SE-specific.

    @classmethod
    @override
    def from_rotation_and_translation(
        cls,
        rotation: SO2,
        translation: hints.Array,
    ) -> "SE2":
        assert translation.shape[-1:] == (2,)
        rotation, translation = broadcast_leading_axes((rotation, translation))
        return SE2(
            unit_complex_xy=jnp.concatenate(
                [rotation.unit_complex, translation], axis=-1
            )
        )

    @override
    def rotation(self) -> SO2:
        return SO2(unit_complex=self.unit_complex_xy[..., :2])

    @override
    def translation(self) -> jax.Array:
        return self.unit_complex_xy[..., 2:]

    # Factory.

    @classmethod
    @override
    def identity(cls, batch_axes: jdc.Static[Tuple[int, ...]] = ()) -> "SE2":
        return SE2(
            unit_complex_xy=jnp.broadcast_to(
                jnp.array([1.0, 0.0, 0.0, 0.0]), (*batch_axes, 4)
            )
        )

    @classmethod
    @override
    def from_matrix(cls, matrix: hints.Array) -> "SE2":
        assert matrix.shape[-2:] == (3, 3)
        # Currently assumes bottom row is [0, 0, 1].
        return SE2.from_rotation_and_translation(
            rotation=SO2.from_matrix(matrix[..., :2, :2]),
            translation=matrix[..., :2, 2],
        )

    # Accessors.

    @override
    def parameters(self) -> jax.Array:
        return self.unit_complex_xy

    @override
    def as_matrix(self) -> jax.Array:
        cos, sin, x, y = jnp.moveaxis(self.unit_complex_xy, -1, 0)
        out = jnp.stack(
            [
                cos,
                -sin,
                x,
                sin,
                cos,
                y,
                jnp.zeros_like(x),
                jnp.zeros_like(x),
                jnp.ones_like(x),
            ],
            axis=-1,
        ).reshape((*self.get_batch_axes(), 3, 3))
        return out

    # Operations.

    @classmethod
    @override
    def exp(cls, tangent: hints.Array) -> "SE2":
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se2.hpp#L558
        # Also see:
        # > http://ethaneade.com/lie.pdf

        assert tangent.shape[-1:] == (3,)

        theta = tangent[..., 2]
        use_taylor = jnp.abs(theta) < get_epsilon(tangent.dtype)

        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        safe_theta = cast(
            jax.Array,
            jnp.where(
                use_taylor,
                jnp.ones_like(theta),  # Any non-zero value should do here.
                theta,
            ),
        )

        theta_sq = theta**2
        sin_over_theta = cast(
            jax.Array,
            jnp.where(
                use_taylor,
                1.0 - theta_sq / 6.0,
                jnp.sin(safe_theta) / safe_theta,
            ),
        )
        one_minus_cos_over_theta = cast(
            jax.Array,
            jnp.where(
                use_taylor,
                0.5 * theta - theta * theta_sq / 24.0,
                (1.0 - jnp.cos(safe_theta)) / safe_theta,
            ),
        )

        V = jnp.stack(
            [
                sin_over_theta,
                -one_minus_cos_over_theta,
                one_minus_cos_over_theta,
                sin_over_theta,
            ],
            axis=-1,
        ).reshape((*tangent.shape[:-1], 2, 2))
        return SE2.from_rotation_and_translation(
            rotation=SO2.from_radians(theta),
            translation=jnp.einsum("...ij,...j->...i", V, tangent[..., :2]),
        )

    @override
    def log(self) -> jax.Array:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se2.hpp#L160
        # Also see:
        # > http://ethaneade.com/lie.pdf

        theta = self.rotation().log()[..., 0]

        cos = jnp.cos(theta)
        cos_minus_one = cos - 1.0
        half_theta = theta / 2.0
        use_taylor = jnp.abs(cos_minus_one) < get_epsilon(theta.dtype)

        # Shim to avoid NaNs in jnp.where branches, which cause failures for
        # reverse-mode AD.
        safe_cos_minus_one = jnp.where(
            use_taylor,
            jnp.ones_like(cos_minus_one),  # Any non-zero value should do here.
            cos_minus_one,
        )

        half_theta_over_tan_half_theta = jnp.where(
            use_taylor,
            # Taylor approximation.
            1.0 - theta**2 / 12.0,
            # Default.
            -(half_theta * jnp.sin(theta)) / safe_cos_minus_one,
        )

        V_inv = jnp.stack(
            [
                half_theta_over_tan_half_theta,
                half_theta,
                -half_theta,
                half_theta_over_tan_half_theta,
            ],
            axis=-1,
        ).reshape((*theta.shape, 2, 2))

        tangent = jnp.concatenate(
            [
                jnp.einsum("...ij,...j->...i", V_inv, self.translation()),
                theta[..., None],
            ],
            axis=-1,
        )
        return tangent

    @override
    def adjoint(self: "SE2") -> jax.Array:
        cos, sin, x, y = jnp.moveaxis(self.unit_complex_xy, -1, 0)
        return jnp.stack(
            [
                cos,
                -sin,
                y,
                sin,
                cos,
                -x,
                jnp.zeros_like(x),
                jnp.zeros_like(x),
                jnp.ones_like(x),
            ],
            axis=-1,
        ).reshape((*self.get_batch_axes(), 3, 3))

    @classmethod
    @override
    def sample_uniform(
        cls, key: jax.Array, batch_axes: jdc.Static[Tuple[int, ...]] = ()
    ) -> "SE2":
        key0, key1 = jax.random.split(key)
        return SE2.from_rotation_and_translation(
            rotation=SO2.sample_uniform(key0, batch_axes=batch_axes),
            translation=jax.random.uniform(
                key=key1,
                shape=(
                    *batch_axes,
                    2,
                ),
                minval=-1.0,
                maxval=1.0,
            ),
        )



"""Tests with explicit examples."""

import jaxlie
import numpy as onp
from hypothesis import given, settings
from hypothesis import strategies as st

from utils import assert_arrays_close, assert_transforms_close, sample_transform


@settings(deadline=None)
@given(_random_module=st.random_module())
def test_se2_translation(_random_module):
    """Simple test for SE(2) translation terms."""
    translation = onp.random.randn(2)
    T = jaxlie.SE2.from_xy_theta(*translation, theta=0.0)
    assert_arrays_close(T @ translation, translation * 2)


@settings(deadline=None)
@given(_random_module=st.random_module())
def test_se3_translation(_random_module):
    """Simple test for SE(3) translation terms."""
    translation = onp.random.randn(3)
    T = jaxlie.SE3.from_rotation_and_translation(
        rotation=jaxlie.SO3.identity(),
        translation=translation,
    )
    assert_arrays_close(T @ translation, translation * 2)


def test_se2_rotation():
    """Simple test for SE(2) rotation terms."""
    T_w_b = jaxlie.SE2.from_rotation_and_translation(
        rotation=jaxlie.SO2.from_radians(onp.pi / 2.0),
        translation=onp.zeros(2),
    )
    p_b = onp.array([1.0, 0.0])
    p_w = onp.array([0.0, 1.0])
    assert_arrays_close(T_w_b @ p_b, p_w)


def test_se3_rotation():
    """Simple test for SE(3) rotation terms."""
    T_w_b = jaxlie.SE3.from_rotation_and_translation(
        rotation=jaxlie.SO3.from_rpy_radians(onp.pi / 2.0, 0.0, 0.0),
        translation=onp.zeros(3),
    )
    T_w_b_alt = jaxlie.SE3.from_rotation(
        jaxlie.SO3.from_rpy_radians(onp.pi / 2.0, 0.0, 0.0),
    )
    p_b = onp.array([0.0, 1.0, 0.0])
    p_w = onp.array([0.0, 0.0, 1.0])
    assert_arrays_close(T_w_b @ p_b, T_w_b_alt @ p_b, p_w)


def test_se3_from_translation():
    """Simple test for SE(3) rotation terms."""
    T_w_b = jaxlie.SE3.from_rotation_and_translation(
        rotation=jaxlie.SO3.identity(),
        translation=onp.arange(3) * 1.0,
    )
    T_w_b_alt = jaxlie.SE3.from_translation(onp.arange(3) * 1.0)
    p_b = onp.array([0.0, 1.0, 0.0])
    p_w = onp.array([0.0, 2.0, 2.0])
    assert_arrays_close(T_w_b @ p_b, T_w_b_alt @ p_b, p_w)


def test_so3_xyzw_basic():
    """Check that we can create an SO3 object from an xyzw quaternion."""
    assert_transforms_close(
        jaxlie.SO3.from_quaternion_xyzw(onp.array([0.0, 0.0, 0.0, 1.0])),
        jaxlie.SO3.identity(),
    )


# def test_so3_xyzw_dtype_error():
#     """Check that an incorrect data-type results in an AssertionError."""
#     with pytest.raises(AssertionError):
#         jaxlie.SO3(onp.array([1, 0, 0, 0])),
#
#
# def test_so3_xyzw_shape_error():
#     """Check that an incorrect shape results in an AssertionError."""
#     with pytest.raises(AssertionError):
#         jaxlie.SO3(onp.array([1.0, 0.0, 0.0, 0.0, 0.0]))


@settings(deadline=None)
@given(_random_module=st.random_module())
def test_se3_compose(_random_module):
    """Compare SE3 composition in matrix form vs compact form."""
    T1 = sample_transform(jaxlie.SE3)
    T2 = sample_transform(jaxlie.SE3)
    assert_arrays_close(T1.as_matrix() @ T2.as_matrix(), (T1 @ T2).as_matrix())
    assert_transforms_close(
        jaxlie.SE3.from_matrix(T1.as_matrix() @ T2.as_matrix()), T1 @ T2
    )



from . import hints as hints
from . import manifold as manifold
from . import utils as utils
from ._base import MatrixLieGroup as MatrixLieGroup
from ._base import SEBase as SEBase
from ._base import SOBase as SOBase
from ._se2 import SE2 as SE2
from ._se3 import SE3 as SE3
from ._so2 import SO2 as SO2
from ._so3 import SO3 as SO3



"""Test manifold helpers."""

from typing import Tuple, Type

import jax
import numpy as onp
import pytest
from jax import numpy as jnp
from jax import tree_util
from utils import (
    assert_arrays_close,
    assert_transforms_close,
    general_group_test,
    general_group_test_faster,
    sample_transform,
)

import jaxlie


@general_group_test
def test_rplus_rminus(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check rplus and rminus on random inputs."""
    T_wa = sample_transform(Group, batch_axes)
    T_wb = sample_transform(Group, batch_axes)
    T_ab = T_wa.inverse() @ T_wb

    assert_transforms_close(jaxlie.manifold.rplus(T_wa, T_ab.log()), T_wb)
    assert_arrays_close(jaxlie.manifold.rminus(T_wa, T_wb), T_ab.log())


@general_group_test
def test_rplus_jacobian(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check analytical rplus Jacobian.."""
    T_wa = sample_transform(Group, batch_axes)

    J_ours = jaxlie.manifold.rplus_jacobian_parameters_wrt_delta(T_wa)

    if batch_axes == ():
        J_jacfwd = _rplus_jacobian_parameters_wrt_delta(T_wa)
        assert_arrays_close(J_ours, J_jacfwd)
    else:
        # Batch axes should match vmap.
        jacfunc = jaxlie.manifold.rplus_jacobian_parameters_wrt_delta
        for _ in batch_axes:
            jacfunc = jax.vmap(jacfunc)
        J_vmap = jacfunc(T_wa)
        assert_arrays_close(J_ours, J_vmap)


@jax.jit
def _rplus_jacobian_parameters_wrt_delta(
    transform: jaxlie.MatrixLieGroup,
) -> jax.Array:
    # Copied from docstring for `rplus_jacobian_parameters_wrt_delta()`.
    return jax.jacfwd(
        lambda delta: jaxlie.manifold.rplus(transform, delta).parameters()
    )(onp.zeros(transform.tangent_dim))


@general_group_test_faster
def test_sgd(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    def loss(transform: jaxlie.MatrixLieGroup):
        return (transform.log() ** 2).sum()

    transform = Group.exp(sample_transform(Group, batch_axes).log())
    original_loss = loss(transform)

    @jax.jit
    def step(t):
        return jaxlie.manifold.rplus(t, -1e-3 * jaxlie.manifold.grad(loss)(t))

    for i in range(5):
        transform = step(transform)

    assert loss(transform) < original_loss


def test_rplus_euclidean():
    assert_arrays_close(
        jaxlie.manifold.rplus(jnp.ones(2), jnp.ones(2)), 2 * jnp.ones(2)
    )


def test_rminus_auto_vmap():
    deltas = jaxlie.manifold.rminus(
        tree_util.tree_map(
            lambda *args: jnp.stack(args),
            [jaxlie.SE3.sample_uniform(jax.random.PRNGKey(0)), jaxlie.SE3.identity()],
        ),
        tree_util.tree_map(
            lambda *args: jnp.stack(args),
            [jaxlie.SE3.identity(), jaxlie.SE3.sample_uniform(jax.random.PRNGKey(0))],
        ),
    )
    assert_arrays_close(deltas[0], -deltas[1])


def test_normalize():
    container = {"key": (jaxlie.SO3(jnp.array([2.0, 0.0, 0.0, 0.0])),)}
    container_valid = {"key": (jaxlie.SO3(jnp.array([1.0, 0.0, 0.0, 0.0])),)}
    with pytest.raises(AssertionError):
        assert_transforms_close(container["key"][0], container_valid["key"][0])
    assert_transforms_close(
        jaxlie.manifold.normalize_all(container)["key"][0], container_valid["key"][0]
    )



"""Tests for group axioms.

https://proofwiki.org/wiki/Definition:Group_Axioms
"""

from typing import Tuple, Type

import numpy as onp
from utils import (
    assert_arrays_close,
    assert_transforms_close,
    general_group_test,
    sample_transform,
)

import jaxlie


@general_group_test
def test_closure(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check closure property."""
    transform_a = sample_transform(Group, batch_axes)
    transform_b = sample_transform(Group, batch_axes)

    composed = transform_a @ transform_b
    assert_transforms_close(composed, composed.normalize())
    composed = transform_b @ transform_a
    assert_transforms_close(composed, composed.normalize())
    composed = Group.multiply(transform_a, transform_b)
    assert_transforms_close(composed, composed.normalize())
    composed = Group.multiply(transform_b, transform_a)
    assert_transforms_close(composed, composed.normalize())


@general_group_test
def test_identity(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check identity property."""
    transform = sample_transform(Group, batch_axes)
    identity = Group.identity(batch_axes)
    assert_transforms_close(transform, identity @ transform)
    assert_transforms_close(transform, transform @ identity)
    assert_arrays_close(
        transform.as_matrix(),
        onp.einsum("...ij,...jk->...ik", identity.as_matrix(), transform.as_matrix()),
    )
    assert_arrays_close(
        transform.as_matrix(),
        onp.einsum("...ij,...jk->...ik", transform.as_matrix(), identity.as_matrix()),
    )


@general_group_test
def test_inverse(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check inverse property."""
    transform = sample_transform(Group, batch_axes)
    identity = Group.identity(batch_axes)
    assert_transforms_close(identity, transform @ transform.inverse())
    assert_transforms_close(identity, transform.inverse() @ transform)
    assert_transforms_close(identity, Group.multiply(transform, transform.inverse()))
    assert_transforms_close(identity, Group.multiply(transform.inverse(), transform))
    assert_arrays_close(
        onp.broadcast_to(
            onp.eye(Group.matrix_dim), (*batch_axes, Group.matrix_dim, Group.matrix_dim)
        ),
        onp.einsum(
            "...ij,...jk->...ik",
            transform.as_matrix(),
            transform.inverse().as_matrix(),
        ),
    )
    assert_arrays_close(
        onp.broadcast_to(
            onp.eye(Group.matrix_dim), (*batch_axes, Group.matrix_dim, Group.matrix_dim)
        ),
        onp.einsum(
            "...ij,...jk->...ik",
            transform.inverse().as_matrix(),
            transform.as_matrix(),
        ),
    )


@general_group_test
def test_associative(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check associative property."""
    transform_a = sample_transform(Group, batch_axes)
    transform_b = sample_transform(Group, batch_axes)
    transform_c = sample_transform(Group, batch_axes)
    assert_transforms_close(
        (transform_a @ transform_b) @ transform_c,
        transform_a @ (transform_b @ transform_c),
    )



from typing import NamedTuple, Union

import jax
import numpy as onp

# Type aliases for JAX/Numpy arrays; primarily for function inputs.

Array = Union[onp.ndarray, jax.Array]
"""Type alias for `Union[jax.Array, onp.ndarray]`."""

Scalar = Union[float, Array]
"""Type alias for `Union[float, Array]`."""


class RollPitchYaw(NamedTuple):
    """Tuple containing roll, pitch, and yaw Euler angles."""

    roll: Scalar
    pitch: Scalar
    yaw: Scalar


__all__ = [
    "Array",
    "Scalar",
    "RollPitchYaw",
]



"""Compare forward- and reverse-mode Jacobians with a numerical Jacobian."""

from functools import lru_cache
from typing import Callable, Tuple, Type, cast

import jax
import numpy as onp
from jax import numpy as jnp
from utils import assert_arrays_close, general_group_test, jacnumerical

import jaxlie

# We cache JITed Jacobians to improve runtime.
cached_jacfwd = lru_cache(maxsize=None)(
    lambda f: jax.jit(jax.jacfwd(f, argnums=1), static_argnums=0)
)
cached_jacrev = lru_cache(maxsize=None)(
    lambda f: jax.jit(jax.jacrev(f, argnums=1), static_argnums=0)
)
cached_jit = lru_cache(maxsize=None)(jax.jit)


def _assert_jacobians_close(
    Group: Type[jaxlie.MatrixLieGroup],
    f: Callable[[Type[jaxlie.MatrixLieGroup], jax.Array], jax.Array],
    primal: jaxlie.hints.Array,
) -> None:
    jacobian_fwd = cached_jacfwd(f)(Group, primal)
    jacobian_rev = cached_jacrev(f)(Group, primal)
    jacobian_numerical = jacnumerical(
        lambda primal: cached_jit(f, static_argnums=0)(Group, primal)
    )(primal)

    assert_arrays_close(jacobian_fwd, jacobian_rev)
    assert_arrays_close(jacobian_fwd, jacobian_numerical, rtol=5e-4, atol=5e-4)


# Exp tests.
def _exp(Group: Type[jaxlie.MatrixLieGroup], generator: jax.Array) -> jax.Array:
    return cast(jax.Array, Group.exp(generator).parameters())


def test_so3_nan():
    """Make sure we don't get NaNs from division when w == 0.

    https://github.com/brentyi/jaxlie/issues/9"""

    @jax.jit
    @jax.grad
    def func(x):
        return jaxlie.SO3.exp(x).log().sum()

    for omega in jnp.eye(3) * jnp.pi:
        a = jnp.array(omega, dtype=jnp.float32)
        assert all(onp.logical_not(onp.isnan(func(a))))


@general_group_test
def test_exp_random(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check that exp Jacobians are consistent, with randomly sampled transforms."""
    del batch_axes  # Not used for autodiff tests.
    generator = onp.random.randn(Group.tangent_dim)
    _assert_jacobians_close(Group=Group, f=_exp, primal=generator)


@general_group_test
def test_exp_identity(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check that exp Jacobians are consistent, with transforms close to the
    identity."""
    del batch_axes  # Not used for autodiff tests.
    generator = onp.random.randn(Group.tangent_dim) * 1e-6
    _assert_jacobians_close(Group=Group, f=_exp, primal=generator)


# Log tests.
def _log(Group: Type[jaxlie.MatrixLieGroup], params: jax.Array) -> jax.Array:
    return Group.log(Group(params))


@general_group_test
def test_log_random(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check that log Jacobians are consistent, with randomly sampled transforms."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim)).parameters()
    _assert_jacobians_close(Group=Group, f=_log, primal=params)


@general_group_test
def test_log_identity(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check that log Jacobians are consistent, with transforms close to the
    identity."""
    params = Group.exp(onp.random.randn(Group.tangent_dim) * 1e-6).parameters()
    _assert_jacobians_close(Group=Group, f=_log, primal=params)


# Adjoint tests.
def _adjoint(Group: Type[jaxlie.MatrixLieGroup], params: jax.Array) -> jax.Array:
    return cast(jax.Array, Group(params).adjoint().flatten())


@general_group_test
def test_adjoint_random(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that adjoint Jacobians are consistent, with randomly sampled transforms."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim)).parameters()
    _assert_jacobians_close(Group=Group, f=_adjoint, primal=params)


@general_group_test
def test_adjoint_identity(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that adjoint Jacobians are consistent, with transforms close to the
    identity."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim) * 1e-6).parameters()
    _assert_jacobians_close(Group=Group, f=_adjoint, primal=params)


# Apply tests.
def _apply(Group: Type[jaxlie.MatrixLieGroup], params: jax.Array) -> jax.Array:
    return Group(params) @ onp.ones(Group.space_dim)


@general_group_test
def test_apply_random(Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]):
    """Check that apply Jacobians are consistent, with randomly sampled transforms."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim)).parameters()
    _assert_jacobians_close(Group=Group, f=_apply, primal=params)


@general_group_test
def test_apply_identity(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that apply Jacobians are consistent, with transforms close to the
    identity."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim) * 1e-6).parameters()
    _assert_jacobians_close(Group=Group, f=_apply, primal=params)


# Multiply tests.
def _multiply(Group: Type[jaxlie.MatrixLieGroup], params: jax.Array) -> jax.Array:
    return cast(jax.Array, (Group(params) @ Group(params)).parameters())


@general_group_test
def test_multiply_random(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that multiply Jacobians are consistent, with randomly sampled
    transforms."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim)).parameters()
    _assert_jacobians_close(Group=Group, f=_multiply, primal=params)


@general_group_test
def test_multiply_identity(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that multiply Jacobians are consistent, with transforms close to the
    identity."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim) * 1e-6).parameters()
    _assert_jacobians_close(Group=Group, f=_multiply, primal=params)


# Inverse tests.
def _inverse(Group: Type[jaxlie.MatrixLieGroup], params: jax.Array) -> jax.Array:
    return cast(jax.Array, Group(params).inverse().parameters())


@general_group_test
def test_inverse_random(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that inverse Jacobians are consistent, with randomly sampled transforms."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim)).parameters()
    _assert_jacobians_close(Group=Group, f=_inverse, primal=params)


@general_group_test
def test_inverse_identity(
    Group: Type[jaxlie.MatrixLieGroup], batch_axes: Tuple[int, ...]
):
    """Check that inverse Jacobians are consistent, with transforms close to the
    identity."""
    del batch_axes  # Not used for autodiff tests.
    params = Group.exp(onp.random.randn(Group.tangent_dim) * 1e-6).parameters()
    _assert_jacobians_close(Group=Group, f=_inverse, primal=params)



"""Helpers for recursively applying tangent-space deltas."""

from typing import Any, Callable, TypeVar, Union, cast, overload

import jax
import numpy as onp
from jax import numpy as jnp

from .. import hints
from .._base import MatrixLieGroup
from .._se2 import SE2
from .._se3 import SE3
from .._so2 import SO2
from .._so3 import SO3
from . import _tree_utils

PytreeType = TypeVar("PytreeType")
GroupType = TypeVar("GroupType", bound=MatrixLieGroup)


def _rplus(transform: GroupType, delta: jax.Array) -> GroupType:
    assert isinstance(transform, MatrixLieGroup)
    assert isinstance(delta, (jax.Array, onp.ndarray))
    return transform @ type(transform).exp(delta)


@overload
def rplus(
    transform: GroupType,
    delta: hints.Array,
) -> GroupType: ...


@overload
def rplus(
    transform: PytreeType,
    delta: _tree_utils.TangentPytree,
) -> PytreeType: ...


# Using our typevars in the overloaded signature will cause errors.
def rplus(
    transform: Union[MatrixLieGroup, Any],
    delta: Union[hints.Array, Any],
) -> Union[MatrixLieGroup, Any]:
    """Manifold right plus. Computes `T' = T @ exp(delta)`.

    Supports pytrees containing Lie group instances recursively; simple Euclidean
    addition will be performed for all other arrays.
    """
    return _tree_utils._map_group_trees(_rplus, jnp.add, transform, delta)


def _rminus(a: GroupType, b: GroupType) -> jax.Array:
    assert isinstance(a, MatrixLieGroup) and isinstance(b, MatrixLieGroup)
    return (a.inverse() @ b).log()


@overload
def rminus(a: GroupType, b: GroupType) -> jax.Array: ...


@overload
def rminus(a: PytreeType, b: PytreeType) -> _tree_utils.TangentPytree: ...


# Using our typevars in the overloaded signature will cause errors.
def rminus(
    a: Union[MatrixLieGroup, Any], b: Union[MatrixLieGroup, Any]
) -> Union[jax.Array, _tree_utils.TangentPytree]:
    """Manifold right minus. Computes
    `delta = T_ab.log() = (T_wa.inverse() @ T_wb).log()`.

    Supports pytrees containing Lie group instances recursively; simple Euclidean
    subtraction will be performed for all other arrays.
    """
    return _tree_utils._map_group_trees(_rminus, jnp.subtract, a, b)


@jax.jit
def rplus_jacobian_parameters_wrt_delta(transform: MatrixLieGroup) -> jax.Array:
    """Analytical Jacobians for `jaxlie.manifold.rplus()`, linearized around a zero
    local delta.

    Mostly useful for reducing JIT compile times for tangent-space optimization.

    Equivalent to --
    ```
    def rplus_jacobian_parameters_wrt_delta(transform: MatrixLieGroup) -> jax.Array:
        # Since transform objects are pytree containers, note that `jacfwd` returns a
        # transformation object itself and that the Jacobian terms corresponding to the
        # parameters are grabbed explicitly.
        return jax.jacfwd(
            jaxlie.manifold.rplus,  # Args are (transform, delta)
            argnums=1,  # Jacobian wrt delta
        )(transform, onp.zeros(transform.tangent_dim)).parameters()
    ```

    Args:
        transform: Transform to linearize around.

    Returns:
        Jacobian. Shape should be `(Group.parameters_dim, Group.tangent_dim)`.
    """
    if isinstance(transform, SO2):
        # Jacobian row indices: cos, sin
        # Jacobian col indices: theta

        J = jnp.zeros((*transform.get_batch_axes(), 2, 1))
        cos, sin = jnp.moveaxis(transform.unit_complex, -1, 0)
        J = J.at[..., 0, 0].set(-sin).at[..., 1, 0].set(cos)

    elif isinstance(transform, SE2):
        # Jacobian row indices: cos, sin, x, y
        # Jacobian col indices: vx, vy, omega
        J = jnp.zeros((*transform.get_batch_axes(), 4, 3))

        # Translation terms.
        J = J.at[..., 2:, :2].set(transform.rotation().as_matrix())

        # Rotation terms.
        J = J.at[..., :2, 2:3].set(
            rplus_jacobian_parameters_wrt_delta(transform.rotation())
        )

    elif isinstance(transform, SO3):
        # Jacobian row indices: qw, qx, qy, qz
        # Jacobian col indices: omega x, omega y, omega z
        w, x, y, z = jnp.moveaxis(transform.wxyz, -1, 0)
        neg_x = -x
        neg_y = -y
        neg_z = -z

        J = (
            jnp.stack(
                [
                    neg_x,
                    neg_y,
                    neg_z,
                    w,
                    neg_z,
                    y,
                    z,
                    w,
                    neg_x,
                    neg_y,
                    x,
                    w,
                ],
                axis=-1,
            ).reshape((*transform.get_batch_axes(), 4, 3))
            / 2.0
        )

    elif isinstance(transform, SE3):
        # Jacobian row indices: qw, qx, qy, qz, x, y, z
        # Jacobian col indices: vx, vy, vz, omega x, omega y, omega z
        J = jnp.zeros((*transform.get_batch_axes(), 7, 6))

        # Translation terms.
        J = J.at[..., 4:, :3].set(transform.rotation().as_matrix())

        # Rotation terms.
        J = J.at[..., :4, 3:6].set(
            rplus_jacobian_parameters_wrt_delta(transform.rotation())
        )

    else:
        assert False, f"Unsupported type: {type(transform)}"

    assert J.shape == (
        *transform.get_batch_axes(),
        transform.parameters_dim,
        transform.tangent_dim,
    )
    return J



from typing import Any, Callable, List, TypeVar

import jax
import numpy as onp
from jax._src.tree_util import _registry  # Dangerous!

from .._base import MatrixLieGroup

# Tangent structures are difficult to annotate, so we just mark everything via Any.
#
# An annotation that would work in most cases is:
#
#     def zero_tangents(structure: T) -> T
#
# But this is leaky; note that an input of List[SE3] should output List[jax.Array],
# Dict[str, SE3] should output Dict[str, SE3], etc.
#
# Another tempting option is to define a wrapper class:
#
#     @jdc.pytree_dataclass
#     class TangentPytree(Generic[PytreeType]):
#         wrapped: Any
#
# And have zero_tangents() return:
#
#     def zero_tangents(structure: T) -> TangentPytree[T]
#
# which we could also use to make `jaxlie.manifold.rplus()` type safe by adding
# overloads to make sure that the delta input is a TangentPytree, but it would be hard
# to accurately annotate the `grad()` and `value_and_grad()` functions with this wrapper
# type without sacrificing the ability to use them as drop-in replacements for
# `jax.grad()` and `jax.value_and_grad()`.
#
# Finally, NewType is also attractive:
#
#     TangentPytree: TypeAlias = NewType("TangentPytree", object)
#
# This seems reasonable, but doesn't play nice with how optax currently (a) annotates
# everything using chex.ArrayTree and (b) doesn't use any generics, leading to a mess of
# casts and `type: ignore` directives. We might consider using this if optax's gradient
# transform annotations change.
TangentPytree = Any


def _map_group_trees(
    f_lie_groups: Callable,
    f_other_arrays: Callable,
    *tree_args,
) -> Any:
    if isinstance(tree_args[0], MatrixLieGroup):
        return f_lie_groups(*tree_args)
    elif isinstance(tree_args[0], (jax.Array, onp.ndarray)):
        return f_other_arrays(*tree_args)
    else:
        # Handle PyTrees recursively.
        assert len(set(map(type, tree_args))) == 1
        registry_entry = _registry[type(tree_args[0])]  # type: ignore

        children: List[List[Any]] = []
        metadata: List[Any] = []
        for tree in tree_args:
            childs, meta = registry_entry.to_iter(tree)
            children.append(childs)
            metadata.append(meta)

        assert len(set(metadata)) == 1

        return registry_entry.from_iter(
            metadata[0],
            [
                _map_group_trees(
                    f_lie_groups,
                    f_other_arrays,
                    *list(children[i][j] for i in range(len(children))),
                )
                for j in range(len(children[0]))
            ],
        )


PytreeType = TypeVar("PytreeType")


def normalize_all(pytree: PytreeType) -> PytreeType:
    """Call `.normalize()` on each Lie group instance in a pytree.

    Results in a naive projection of each group instance to its respective manifold.
    """

    def _project(t: MatrixLieGroup) -> MatrixLieGroup:
        return t.normalize()

    return _map_group_trees(
        _project,
        lambda x: x,
        pytree,
    )



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="jaxlie",
    version="1.4.1",
    description="Matrix Lie groups in JAX",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/brentyi/jaxlie",
    author="brentyi",
    author_email="brentyi@berkeley.edu",
    license="MIT",
    packages=find_packages(),
    package_data={"jaxlie": ["py.typed"]},
    python_requires=">=3.8",
    install_requires=[
        "jax>=0.3.18",  # For jax.Array.
        "jax_dataclasses>=1.4.4",
        "numpy",
        "typing_extensions>=4.0.0",
        "tyro",  # Only used in examples.
    ],
    extras_require={
        "testing": [
            "mypy",
            # https://github.com/google/jax/issues/12536
            "jax!=0.3.19",
            "flax",
            "hypothesis[numpy]",
            "pytest",
            "pytest-xdist[psutil]",
            "pytest-cov",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
